@@ -8,8 +8,8 @@ import ujson as json
 import os
 import asyncio
 from aiofiles import open as aopen
-import aiofiles.os as aos
-from aiofiles.threadpool.text import AsyncTextIOWrapper as AWrapper
+from aiofiles.threadpool.text import AsyncTextIOWrapper as TWrapper
+from aiofiles.threadpool.binary import AsyncFileIO as BWrapper
 from .event_hooks import EventHook, AsyncActionChain, EventHint
 from abc import ABC, abstractmethod
 from typing import Any, Callable, Awaitable, TypeVar
@@ -35,15 +35,6 @@ def touch(path: str, create_dirs: bool):
     with open(path, 'a'):
         pass
 
-class StorageEventHook(EventHook):
-    def __init__(self):
-        super().__init__()
-        self.hook('write.pre', AsyncActionChain())
-        self.hook('write.post', AsyncActionChain(limit=1))
-        self.hook('read.pre', AsyncActionChain(limit=1))
-        self.hook('read.post', AsyncActionChain())
-        self.hook('close', AsyncActionChain())
-
 
 class Storage(ABC):
     """
@@ -58,8 +49,8 @@ class Storage(ABC):
 
     def __init__(self):
         # Create event hook
-        self._event_hook = StorageEventHook()
-        self._on = StorageHints(self._event_hook)
+        self._event_hook = EventHook()
+        self._on = EventHint(self._event_hook)
 
     @property
     def on(self) -> StorageHints:
@@ -69,12 +60,21 @@ class Storage(ABC):
         return self._on
 
     @property
-    def event_hook(self) -> StorageEventHook:
+    def event_hook(self) -> EventHook:
         """
         The event hook for this storage.
         """
         return self._event_hook
 
+    @property
+    @abstractmethod
+    def closed(self) -> bool:
+        """
+        Whether the storage is closed.
+        """
+
+        pass
+
     @abstractmethod
     async def read(self) -> dict[str, dict[str, Any]] | None:
         """
@@ -134,10 +134,22 @@ class JSONStorage(Storage):
             touch(path, create_dirs=create_dirs)
 
         # Open the file for reading/writing
-        self._handle: AWrapper | None = None
+        self._handle: TWrapper|BWrapper|None = None
         self._path = path
         self._encoding = encoding
 
+        # Initialize event hooks
+        self.event_hook.hook('write.pre', AsyncActionChain())
+        self.event_hook.hook('write.post', AsyncActionChain(limit=1))
+        self.event_hook.hook('read.pre', AsyncActionChain(limit=1))
+        self.event_hook.hook('read.post', AsyncActionChain())
+        self.event_hook.hook('close', AsyncActionChain())
+        self._on = StorageHints(self._event_hook)  # Add hints for event hooks
+
+    @property
+    def closed(self) -> bool:
+        return self._handle is not None and self._handle.closed
+
     async def close(self) -> None:
         await self._event_hook.aemit('close', self)
         if self._handle is not None:
@@ -148,6 +160,8 @@ class JSONStorage(Storage):
             self._handle = await aopen(self._path, self._mode, encoding=self._encoding)
         # Get the file size by moving the cursor to the file end and reading
         # its location
+        if self._handle.closed:
+            raise IOError('File is closed')
         await self._handle.seek(0, os.SEEK_END)
         size = await self._handle.tell()
 
@@ -165,29 +179,32 @@ class JSONStorage(Storage):
             pre = await self._event_hook.aemit('read.pre', self, raw)
             if pre and pre[0] is not None:
                 raw = pre[0]
-            data = json.loads(raw)
+            data = json.loads(raw or "{}")
             await self._event_hook.aemit('read.post', self, data)
             return data
 
     async def write(self, data: dict[str, dict[str, Any]]):
+        data = data.copy()
         if self._handle is None:
             self._handle = await aopen(self._path, self._mode, encoding=self._encoding)
+        if self._handle.closed:
+            raise IOError('File is closed')
         # Move the cursor to the beginning of the file just in case
         await self._handle.seek(0)
 
         # Trigger write events
         await self._event_hook.aemit('write.pre', self, data)
         # Serialize the database state using the user-provided arguments
-        serialized = json.dumps(data, **self.kwargs)
+        serialized: bytes|str = json.dumps(data or {}, **self.kwargs)
         if 'b' in self._mode:
-            serialized = serialized.encode(self._encoding)
+            serialized = serialized.encode()  # type: ignore
         post = await self._event_hook.aemit('write.post', self, serialized)
         if post and post[0] is not None:  # if action returned something
             serialized = post[0]
 
         # Write the serialized data to the file
         try:
-            await self._handle.write(serialized)
+            await self._handle.write(serialized)  # type: ignore
         except io.UnsupportedOperation:
             raise IOError(f"Cannot write to the database. Access mode is '{self._mode}'")
 
@@ -214,6 +231,10 @@ class MemoryStorage(Storage):
         super().__init__()
         self.memory = None
 
+    @property
+    def closed(self) -> bool:
+        return False
+
     async def read(self) -> dict[str, dict[str, Any]] | None:
         return self.memory
 
@@ -224,38 +245,34 @@ class MemoryStorage(Storage):
 ############# Event Hints #############
 
 _W = TypeVar('_W', bound=Callable[[str, Storage, dict[str, dict[str, Any]]], Awaitable[None]])
-_R = TypeVar('_R', bound=Callable[[str, Storage, str|bytes], Awaitable[str|bytes|None]])
+_R = TypeVar('_R', bound=Callable[[str, Storage, Any], Awaitable[Any|None]])
 _C = TypeVar('_C', bound=Callable[[str, Storage], Awaitable[None]])
 class _write_hint(EventHint):
     @property
     def pre(self) -> Callable[[_W], _W]:
         """Action Type: (event_name: str, Storage, data: dict[str, dict[str, Any]]) -> None"""
-        return self._chain.pre
     @property
     def post(self) -> Callable[[_R], _R]:
         """Action Type: (event_name: str, Storage, data: str|bytes) -> str|bytes|None"""
-        return self._chain.post
 class _read_hint(EventHint):
     @property
     def pre(self) -> Callable[[_R], _R]:
         """Action Type: (event_name: str, Storage, data: str|bytes) -> str|bytes|None"""
-        return self._chain.pre
     @property
     def post(self) -> Callable[[_W], _W]:
         """Action Type: (event_name: str, Storage, data: dict[str, dict[str, Any]]) -> None"""
-        return self._chain.post
 class StorageHints(EventHint):
     """
     Event hints for the storage class.
     """
     @property
     def write(self) -> _write_hint:
-        return self._chain.write
+        return self._chain.write  # type: ignore
     @property
     def read(self) -> _read_hint:
-        return self._chain.read
+        return self._chain.read  # type: ignore
     @property
     def close(self) -> Callable[[_C], _C]:
         return self._chain.close
 
-############# Event Hints #############
\ No newline at end of file
+############# Event Hints #############
