@@ -3,29 +3,120 @@ This module implements tables, the central place for accessing and manipulating
 data in TinyDB.
 """
 
-import asyncio
+from __future__ import annotations
+from abc import ABC, abstractmethod
 from typing import (
+    TYPE_CHECKING,
     AsyncGenerator,
+    overload,
     Callable,
-    Dict,
     Iterable,
-    List,
     Mapping,
-    Optional,
-    Union,
+    Generic,
     cast,
-    Tuple
+    TypeVar,
+    Type,
+    Any
 )
-import nest_asyncio
-nest_asyncio.apply()
 from .queries import QueryLike
 from .storages import Storage
-from .utils import LRUCache
+from .utils import LRUCache, ensure_async, sync_await
 
-__all__ = ('Document', 'Table')
+__all__ = ('Document', 'Table', 'BaseID', 'IncreID', 'BaseDocument')
+IDVar = TypeVar('IDVar', bound="BaseID")
+DocVar = TypeVar('DocVar', bound="BaseDocument")
+
+class BaseID(ABC):
+    @abstractmethod
+    def __init__(self: IDVar, value: str|IDVar):
+        super().__init__()
+    @abstractmethod
+    def __str__(self) -> str:
+        raise NotImplementedError()
+    @abstractmethod
+    def __hash__(self) -> int:
+        raise NotImplementedError()
+    @abstractmethod
+    def __eq__(self, other: object) -> bool:
+        raise NotImplementedError()
+    @classmethod
+    @abstractmethod
+    def next_id(cls, table: Table):
+        """Get the next ID for the given table. Can be defined as an async func."""
+        raise NotImplementedError()
+    @classmethod
+    @abstractmethod
+    def mark_exists(cls, table: Table, new_id):
+        """Mark the given id as existing. Can be defined as an async func."""
+        raise NotImplementedError()
+    @classmethod
+    @abstractmethod
+    def clear_cache(cls, table: Table):
+        """Clear the ID cache for a table. CANNOT be defined as an async func."""
+        raise NotImplementedError()
 
 
-class Document(dict):
+class IncreID(BaseID, int):
+    """ID class using incrementing integers."""
+    _cache: dict[str, int] = {}
+    def __init__(self, value: str|int|IncreID):
+        self._value = int(value)
+    def __str__(self) -> str:
+        return str(self._value)
+    def __int__(self) -> int:
+        return self._value
+    def __hash__(self) -> int:
+        return self._value
+    def __eq__(self, other: object) -> bool:
+        if isinstance(other, IncreID):
+            return self._value == other._value
+        elif isinstance(other, (int, float)):
+            return self._value == other
+        return False
+    @classmethod
+    async def next_id(cls, table: Table) -> IncreID:
+        # If we already know the next ID
+        if table.name in cls._cache:
+            new = cls(cls._cache[table.name])
+            cls._cache[table.name] += 1
+            return new
+
+        # Read the table documents
+        docs = await table._read_table()
+
+        # If the table is empty, set the initial ID
+        if not docs:
+            next_id = 1
+            cls._cache[table.name] = next_id + 1
+            return cls(next_id)
+
+        # Determine the next ID based on the maximum ID that's currently in use
+        max_id = max(int(i) for i in docs.keys())
+        next_id = max_id + 1
+
+        # The next ID we wil return AFTER this call needs to be larger than
+        # the current next ID we calculated
+        cls._cache[table.name] = next_id + 1
+        return cls(next_id)
+    @classmethod
+    def mark_exists(cls, table: Table, new_id: IncreID):
+        cls._cache[table.name] = max(cls._cache.get(table.name, 0), int(new_id) + 1)
+    @classmethod
+    def clear_cache(cls, table: Table):
+        cls._cache.pop(table.name, None)
+
+
+class BaseDocument(Mapping[IDVar, Any]):
+    @property
+    @abstractmethod
+    def doc_id(self) -> IDVar:
+        raise NotImplementedError()
+    @doc_id.setter
+    def doc_id(self, value: IDVar):
+        raise NotImplementedError()
+
+
+class Document(dict, BaseDocument[IDVar]):
     """
     A document stored in the database.
 
@@ -33,12 +124,19 @@ class Document(dict):
     its ID using ``doc.doc_id``.
     """
 
-    def __init__(self, value: Mapping, doc_id: int):
+    def __init__(self, value: Mapping, doc_id: IDVar):
         super().__init__(value)
         self.doc_id = doc_id
 
+    @property
+    def doc_id(self) -> IDVar:
+        return self._doc_id
+    @doc_id.setter
+    def doc_id(self, value: IDVar):
+        self._doc_id = value
 
-class Table:
+
+class Table(Generic[IDVar, DocVar]):
     """
     Represents a single TinyDB table.
 
@@ -79,12 +177,12 @@ class Table:
     #: The class used to represent documents
     #:
     #: .. versionadded:: 4.0
-    document_class = Document
+    # document_class = Document
 
     #: The class used to represent a document ID
     #:
     #: .. versionadded:: 4.0
-    document_id_class = int
+    # document_id_class = IncreID  # moved to __init__
 
     #: The class used for caching query results
     #:
@@ -95,32 +193,52 @@ class Table:
     #:
     #: .. versionadded:: 4.0
     default_query_cache_capacity = 10
-
+    # A stupid workaround for mypy
+    @overload
+    def __init__(self:Table[IncreID, Document], storage: Storage, name: str, 
+                cache_size: int = default_query_cache_capacity): ...
+    @overload
+    def __init__(self: Table[IncreID, DocVar], storage: Storage, name: str,
+        cache_size: int = default_query_cache_capacity,*, document_class: Type[DocVar]): ...
+    @overload
+    def __init__(self: Table[IDVar, Document], storage: Storage, name: str,
+        cache_size: int = default_query_cache_capacity,*, document_id_class: Type[IDVar]): ...
+    @overload
+    def __init__(self: Table[IDVar, DocVar], storage: Storage, name: str,
+        cache_size: int = default_query_cache_capacity,
+        *, document_id_class: Type[IDVar], document_class: Type[DocVar],
+    ):
+        ...
     def __init__(
         self,
         storage: Storage,
         name: str,
-        cache_size: int = default_query_cache_capacity
+        cache_size: int = default_query_cache_capacity,
+        *,
+        document_id_class = IncreID,
+        document_class = Document,
     ):
         """
         Create a table instance.
         """
 
+        self.document_id_class = document_id_class
+        self.document_class = document_class
         self._storage = storage
         self._name = name
-        self._query_cache: LRUCache[QueryLike, List[Document]] \
+        self._query_cache: LRUCache[QueryLike, list[DocVar]] \
             = self.query_cache_class(capacity=cache_size)
 
-        self._next_id = None
+        self.document_id_class.clear_cache(self)  # clear the ID cache
 
     def __repr__(self):
         args = [
-            'name={!r}'.format(self.name),
-            'total={}'.format(len(self)),
-            'storage={}'.format(self._storage),
+            f"name='{self.name}'",
+            f"total={len(self)}",
+            f"storage={self._storage}",
         ]
 
-        return '<{} {}>'.format(type(self).__name__, ', '.join(args))
+        return f"<{type(self).__name__} {', '.join(args)}>"
 
     @property
     def name(self) -> str:
@@ -136,7 +254,7 @@ class Table:
         """
         return self._storage
 
-    async def insert(self, document: Mapping) -> int:
+    async def insert(self, document: Mapping) -> IDVar:
         """
         Insert a new document into the table.
 
@@ -149,13 +267,15 @@ class Table:
             raise ValueError('Document is not a Mapping')
 
         # First, we get the document ID for the new document
-        if isinstance(document, Document):
+        if isinstance(document, self.document_class):
             # For a `Document` object we use the specified ID
             doc_id = document.doc_id
 
             # We also reset the stored next ID so the next insert won't
             # re-use document IDs by accident when storing an old value
-            self._next_id = None
+            if TYPE_CHECKING:
+                assert(isinstance(doc_id, self.document_id_class))
+            await ensure_async(self.document_id_class.mark_exists)(self, doc_id)
         else:
             # In all other cases we use the next free ID
             doc_id = await self._get_next_id()
@@ -176,7 +296,7 @@ class Table:
 
         return doc_id
 
-    async def insert_multiple(self, documents: Iterable[Mapping]) -> List[int]:
+    async def insert_multiple(self, documents: Iterable[Mapping]) -> list[IDVar]:
         """
         Insert multiple documents into the table.
 
@@ -192,7 +312,7 @@ class Table:
                 if not isinstance(document, Mapping):
                     raise ValueError('Document is not a Mapping')
 
-                if isinstance(document, Document):
+                if isinstance(document, self.document_class):
                     # Check if document does not override an existing document
                     if document.doc_id in table:
                         raise ValueError(
@@ -211,8 +331,7 @@ class Table:
                 # Generate new document ID for this document
                 # Store the doc_id, so we can return all document IDs
                 # later, then save the document with the new doc_id
-                loop = asyncio.get_event_loop()
-                doc_id = loop.run_until_complete(self._get_next_id())
+                doc_id = sync_await(self._get_next_id())
                 doc_ids.append(doc_id)
                 table[doc_id] = dict(document)
 
@@ -221,7 +340,7 @@ class Table:
 
         return doc_ids
 
-    async def all(self) -> List[Document]:
+    async def all(self) -> list[DocVar]:
         """
         Get all documents stored in the table.
 
@@ -235,7 +354,7 @@ class Table:
 
         return [i async for i in self]
 
-    async def search(self, cond: QueryLike) -> List[Document]:
+    async def search(self, cond: QueryLike) -> list[DocVar]:
         """
         Search for all documents matching a 'where' cond.
 
@@ -281,9 +400,9 @@ class Table:
 
     async def get(
         self,
-        cond: Optional[QueryLike] = None,
-        doc_id: Optional[int] = None,
-    ) -> Optional[Document]:
+        cond: QueryLike = None,
+        doc_id: IDVar = None,
+    ) -> DocVar|None:
         """
         Get exactly one document specified by a query or a document ID.
 
@@ -325,8 +444,8 @@ class Table:
 
     async def contains(
         self,
-        cond: Optional[QueryLike] = None,
-        doc_id: Optional[int] = None
+        cond: QueryLike = None,
+        doc_id: IDVar = None
     ) -> bool:
         """
         Check whether the database contains a document matching a query or
@@ -349,10 +468,10 @@ class Table:
 
     async def update(
         self,
-        fields: Union[Mapping, Callable[[Mapping], None]],
-        cond: Optional[QueryLike] = None,
-        doc_ids: Optional[Iterable[int]] = None,
-    ) -> List[int]:
+        fields: Mapping|Callable[[Mapping], None],
+        cond: QueryLike = None,
+        doc_ids: Iterable[IDVar] = None,
+    ) -> list[IDVar]:
         """
         Update all matching documents to have a given set of fields.
 
@@ -441,9 +560,9 @@ class Table:
     async def update_multiple(
         self,
         updates: Iterable[
-            Tuple[Union[Mapping, Callable[[Mapping], None]], QueryLike]
+            tuple[Mapping | Callable[[Mapping], None], QueryLike]
         ],
-    ) -> List[int]:
+    ) -> list[IDVar]:
         """
         Update all matching documents to have a given set of fields.
 
@@ -490,7 +609,7 @@ class Table:
 
         return updated_ids
 
-    async def upsert(self, document: Mapping, cond: Optional[QueryLike] = None) -> List[int]:
+    async def upsert(self, document: Mapping, cond: QueryLike = None) -> list[IDVar]:
         """
         Update documents, if they exist, insert them otherwise.
 
@@ -505,8 +624,8 @@ class Table:
         """
 
         # Extract doc_id
-        if isinstance(document, Document) and hasattr(document, 'doc_id'):
-            doc_ids: Optional[List[int]] = [document.doc_id]
+        if isinstance(document, self.document_class) and hasattr(document, 'doc_id'):
+            doc_ids: list[IDVar] | None = [document.doc_id]
         else:
             doc_ids = None
 
@@ -518,7 +637,7 @@ class Table:
 
         # Perform the update operation
         try:
-            updated_docs: Optional[List[int]] = await self.update(document, cond, doc_ids)
+            updated_docs: list[IDVar] | None = await self.update(document, cond, doc_ids)
         except KeyError:
             # This happens when a doc_id is specified, but it's missing
             updated_docs = None
@@ -533,9 +652,9 @@ class Table:
 
     async def remove(
         self,
-        cond: Optional[QueryLike] = None,
-        doc_ids: Optional[Iterable[int]] = None,
-    ) -> List[int]:
+        cond: QueryLike = None,
+        doc_ids: Iterable[IDVar] = None,
+    ) -> list[IDVar]:
         """
         Remove all matching documents.
 
@@ -603,7 +722,7 @@ class Table:
         await self._update_table(lambda table: table.clear())
 
         # Reset document ID counter
-        self._next_id = None
+        self.document_id_class.clear_cache(self)
 
     async def count(self, cond: QueryLike) -> int:
         """
@@ -625,11 +744,10 @@ class Table:
         """
         Count the total number of documents in this table.
         """
-        loop = asyncio.get_event_loop()
-        table = loop.run_until_complete(self._read_table())
+        table = sync_await(self._read_table())
         return len(table)
 
-    async def __aiter__(self) -> AsyncGenerator[Document, Document]:
+    async def __aiter__(self) -> AsyncGenerator[DocVar, DocVar]:
         """
         Iterate over all documents stored in the table.
 
@@ -646,37 +764,13 @@ class Table:
         Return the ID for a newly inserted document.
         """
 
-        # If we already know the next ID
-        if self._next_id is not None:
-            next_id = self._next_id
-            self._next_id = next_id + 1
-
-            return next_id
-
-        # Determine the next document ID by finding out the max ID value
-        # of the current table documents
-
-        # Read the table documents
-        table = await self._read_table()
-
-        # If the table is empty, set the initial ID
-        if not table:
-            next_id = 1
-            self._next_id = next_id + 1
-
-            return next_id
-
-        # Determine the next ID based on the maximum ID that's currently in use
-        max_id = max(self.document_id_class(i) for i in table.keys())
-        next_id = max_id + 1
-
-        # The next ID we wil return AFTER this call needs to be larger than
-        # the current next ID we calculated
-        self._next_id = next_id + 1
+        # Determine the next document ID by calling self.document_class.next_id()
+        # Make sure it doesn't block the event loop
+        next_id = await ensure_async(self.document_id_class.next_id)(self)
 
         return next_id
 
-    async def _read_table(self) -> Dict[str, Mapping]:
+    async def _read_table(self) -> dict[str, Mapping]:
         """
         Read the table data from the underlying storage.
 
@@ -701,7 +795,7 @@ class Table:
 
         return table
 
-    async def _update_table(self, updater: Callable[[Dict[int, Mapping]], None]):
+    async def _update_table(self, updater: Callable[[dict[IDVar, Mapping]], None]):
         """
         Perform a table update operation.
 
